/*
 * Copyright (c) 2003-2005 Erez Zadok
 * Copyright (c) 2003-2005 Charles P. Wright
 * Copyright (c) 2003-2005 Mohammad Nayyer Zubair
 * Copyright (c) 2003-2005 Puja Gupta
 * Copyright (c) 2003-2005 Harikesavan Krishnan
 * Copyright (c) 2003-2005 Stony Brook University
 * Copyright (c) 2003-2005 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */
/*
 *  $Id: subr.c,v 1.96 2005/01/26 16:33:37 cwright Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "unionfs.h"

/* lookup files/directories from index 'bstart' onwards, does not looksup bstart */
STATIC int
unionfs_partial_lookup(dentry_t *dentry)
{
    int err = 0;
    int dentry_count = 0;
    int bstart, bindex;
    const char *name;
    unsigned int namelen;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    int positive;

    print_entry_location();

    fist_print_dentry("IN: unionfs_partial_lookup: ", dentry);
    lock_dpd(dentry);
    bstart = dbstart(dentry);

    name = dentry->d_name.name;
    namelen = dentry->d_name.len;

    positive = dentry->d_inode ? 1 :  0;

    for (bindex = bstart + 1; bindex < sbmax(dentry->d_sb); bindex++) {
	/* Check if there is already something there. */
	hidden_dentry = dtohd_index(dentry, bindex);
	if (hidden_dentry) {
	    continue;
	}

	/* Find out hidden parent dentry */
	hidden_dir_dentry = dtohd_index(dentry->d_parent, bindex);

	/* continue if hidden parent doesn't exist */
	if (!hidden_dir_dentry) {
	    continue;
	}

	/* lookup underlying file system */
	hidden_dentry = lookup_one_len(name, hidden_dir_dentry, namelen);
	if (IS_ERR(hidden_dentry)) {
	    printk("ERR from hidden_dentry!!!\n");
	    err = PTR_ERR(hidden_dentry);
	    goto out;
	}

	if (hidden_dentry->d_inode) {
	    /* this is not a negative dentry, so init dtohd with
	     * this new dentry.  Also, change bend.
	     */
	    set_dtohd_index(dentry, bindex, hidden_dentry);
	    set_dbend(dentry, bindex);
	    dentry_count++;
	} else {
	    /* dput the unnecessary negative dentry */
	    dput(hidden_dentry);
	}
    } // end for

    if (dentry_count) {
	if (positive) {
    		/* need to connect the new inodes to the pre-existing
     		 * unionfs inode, unionfs_reinterpose
      		 */
		unionfs_reinterpose(dentry);
	} else {
		int bstart;
		int bend;

		/* Somehow the file was created for us, so we need to interpose it with a new inode. */

		bstart = dbstart(dentry);
		bend = dbend(dentry);

        	for (bindex = bstart; bindex <= bend; bindex++) {
               		if (dtohd_index(dentry, bindex)) {
				if (!dtohd_index(dentry, bindex)->d_inode) {
					dput(dtohd_index(dentry, bindex));
					set_dtohd_index(dentry, bindex, NULL);
				} else {
					set_dbstart(dentry, bindex);
                        		break;
				}
			}
		}

		unionfs_interpose(dentry, dentry->d_sb, 0);
	}
    }

 out:
    unlock_dpd(dentry);
    fist_print_dentry("OUT: unionfs_partial_lookup: ", dentry);
    print_exit_status(err);
    return err;
}


struct create_dir_whs_callback {
    int err;
    dentry_t *dentry;
    int write_index;
    int filldir_called;
};


int vfs_create_nozombie(struct inode *dir, struct dentry *dentry, int mode);
/* This is designed to be called just like a filldir function. */
static int create_dir_whs_filldir(void *dirent, const char *name, int namelen, loff_t offset, ino_t ino, unsigned int d_type)
{
    int err = 0;
    struct create_dir_whs_callback *buf = (struct create_dir_whs_callback *) dirent;
    dentry_t *dentry;
    int write_index;
    dentry_t *hidden_parent_dentry;
    dentry_t *hidden_dentry;
    dentry_t *hidden_whiteout_dentry;
    char *wh_name = NULL;

    buf->filldir_called = 1;

    /* skip '.' and '..' entries */
    if (name[0] == '.' && (namelen == 1 || (name[1] == '.' && namelen == 2))) {
        goto out;
    }

    wh_name = KMALLOC(sizeof(char) * (namelen + 5), GFP_NOFS);
    if (!wh_name) {
        err = -ENOMEM;
        goto out;
    }
    strcpy(wh_name, ".wh.");
    strncat(wh_name, name, namelen);
    wh_name[4 + namelen] = '\0';

    dentry = buf->dentry;
    write_index = buf->write_index;

    hidden_dentry = dtohd_index(dentry, write_index);
    hidden_whiteout_dentry = lookup_one_len(wh_name, hidden_dentry, namelen + 4);
    if (IS_ERR(hidden_dentry)) {
	/*
	 * this produces an unusual dentry: one that has neither an
	 * inode, nor a private structure attached to it. All cleanup
	 * methods (d_delete, d_release, etc) must be prepared to deal
	 * with such dentries. Ion 09/29/2001
	 */
	printk("ERR from hidden_dentry!!!\n");
	err = PTR_ERR(hidden_dentry);
	goto out;
    }

    if (!hidden_whiteout_dentry->d_inode) {
    	hidden_parent_dentry = lock_parent(hidden_whiteout_dentry);

    	err = vfs_create_nozombie(hidden_parent_dentry->d_inode, hidden_whiteout_dentry, 0666);
    	fist_copy_attr_times(dentry->d_inode, hidden_parent_dentry->d_inode);
    	unlock_dir(hidden_parent_dentry);
    }
    dput(hidden_whiteout_dentry);

 out:
    if (wh_name) {
        KFREE(wh_name);
    }
    buf->err = err;
    return err;
}

/* Creates a whiteout for the files existing in a directory.  Pass the unionfs dentry,
 * and the bindex
 */
int create_dir_whs(dentry_t *dentry, int cur_index)
{
    int err = 0;
    int bindex = 0, bstart, bend;
    dentry_t *hidden_dentry = NULL;
    const char *name;
    unsigned int namelen;
    dentry_t *hidden_parent_dentry;
    super_block_t *sb;
    file_t *hidden_file;
    struct create_dir_whs_callback buf;

    sb = dentry->d_sb;

    name = dentry->d_name.name;
    namelen = dentry->d_name.len;

    bstart = dbstart(dentry);
    bend = dbend(dentry);

    print_entry_location();

    fist_print_dentry("IN create_dir_whs: ", dentry);

    /* open all dentries to the right of branch 'cur_index' and do vfs_readdir on them.
       In the create_dir_whs_filldir() function, for each entry create a whiteout in branch 'cur_index'  */
    for (bindex = cur_index + 1; bindex <= sbend(sb); bindex++) {

	/* Find out hidden parent dentry */
	hidden_parent_dentry = dtohd_index(dentry->d_parent, bindex);

	/* continue if hidden parent doesn't exist */
	if (!hidden_parent_dentry) {
	    continue;
	}

	/* lookup underlying file system */
	hidden_dentry = lookup_one_len(name, hidden_parent_dentry, namelen);
	if (IS_ERR(hidden_dentry)) {
	    err = PTR_ERR(hidden_dentry);
	    goto out;
	}

	if (hidden_dentry->d_inode && S_ISDIR(hidden_dentry->d_inode->i_mode)) {

	    mntget(stohiddenmnt_index(sb, bindex));
	    branchget(sb, bindex);
	    hidden_file = dentry_open(hidden_dentry, stohiddenmnt_index(sb, bindex), O_RDONLY);
	    if (IS_ERR(hidden_file)) {
		err = PTR_ERR(hidden_file);
                branchput(sb, bindex);
		goto out;
	    }
	    /* prepare callback buffer */
	    buf.dentry = dentry;
	    buf.write_index = cur_index;

	    do {
	        buf.filldir_called = 0;
	        err = vfs_readdir(hidden_file, create_dir_whs_filldir, &buf);
	        if (buf.err) {
		    err = buf.err;
	        }
	    } while (!err && buf.filldir_called);

	    /* fput calls dput for internally */
            fput(hidden_file);
            branchput(sb, bindex);

	} else {
	    /* dput the unnecessary negative dentry */
	    dput(hidden_dentry);
	}
    }

 out:
    fist_print_dentry("OUT create_dir_whs :", dentry);
    print_exit_status(err);
    return err;
}

/* Pass an unionfs dentry and an index.  It will try to create a whiteout for the filename
 * in dentry, and will try in branch 'index'.  On error, it will proceed to a branch
 * to the left
 */
int create_whiteout(dentry_t *dentry, int start)
{
    int bstart, bend, bindex;
    dentry_t *hidden_dir_dentry;
    dentry_t *hidden_dentry;
    dentry_t *hidden_wh_dentry;
    char *name = NULL;
    int err = -EUNIONFS_NO_WHITEOUT;

    print_entry_location();

    PASSERT(dentry);
    PASSERT(dentry->d_parent);

    fist_print_dentry("IN create_whiteout", dentry);
    bstart = dbstart(dentry);
    bend = dbend(dentry);

    /* create dentry's whiteout equivalent */
    name = KMALLOC(sizeof(char) * (dentry->d_name.len + 5), GFP_NOFS);
    if (!name) {
        err = -ENOMEM;
        goto out;
    }
    strcpy(name, ".wh.");
    strncat(name, dentry->d_name.name, dentry->d_name.len);
    name[dentry->d_name.len + 4] = '\0';

    /* start from 'start', create foo then rename it to .wh.foo, on error go to
       the next left branch, if hidden_dentry is NULL do recursive directory creation */
    for (bindex = start; bindex >= 0; bindex--) {
	hidden_dentry = dtohd_index(dentry, bindex);

	if (!hidden_dentry) {
	    /* if hidden dentry is not present, create the entire
	     * hidden dentry directory structure and go ahead.
	     * Since we want to just create whiteout, we only want the parent
	     * dentry, and hence get rid of this dentry.
	     */
	    hidden_dentry = unionfs_create_dirs(dentry->d_inode, dentry, bindex);
	    if (!hidden_dentry || IS_ERR(hidden_dentry)) {
		fist_dprint(8, "hidden dentry NULL for bindex = %d\n", bindex);
		continue;
	    }
	}
	hidden_wh_dentry = lookup_one_len(name, hidden_dentry->d_parent, dentry->d_name.len + 4);

	if (!hidden_wh_dentry || IS_ERR(hidden_wh_dentry)) {
	    continue;
	}
	ASSERT(!hidden_wh_dentry->d_inode);

	hidden_dir_dentry = lock_parent(hidden_wh_dentry);
    	if (!(err = is_robranch_super(dentry->d_sb, bindex))) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	    err = vfs_create(hidden_dir_dentry->d_inode, hidden_wh_dentry, ~current->fs->umask & S_IRWXUGO);
#else
	    err = vfs_create(hidden_dir_dentry->d_inode, hidden_wh_dentry, ~current->fs->umask & S_IRWXUGO,NULL);

#endif
	}
	unlock_dir(hidden_dir_dentry);
	dput(hidden_wh_dentry);


	if (!err) {
	    break;
        } else {
            if (IS_SET(dentry->d_sb, GLOBAL_ERR_PASSUP)) {
                goto out;
            }
            /* break out of for loop if error returned was NOT -EROFS */
            if (!IS_COPYUP_ERR(err)) {
                break;
            }
        }
    } // end for

    fist_print_dentry("OUT create_whiteout", dentry);
out:
    if (name) {
       KFREE(name);
    }
    print_exit_status(err);
    return err;
}

/* Pass an unionfs dentry, name of file, and an index.  It will try to create a whiteout
 * for the filename passed in the parent directory passed.  It will start trying
 * in branch start.  On error, it will proceed to a branch to the left
 */
int create_whiteout_parent(dentry_t *parent_dentry, const char *filename, int start)
{
    int bindex;
    int old_bstart, old_bend;
    dentry_t *hidden_dir_dentry;
    dentry_t *hidden_grand_parent_dentry;
    dentry_t *hidden_parent_dentry;
    dentry_t *hidden_wh_dentry;
    char *name = NULL;
    int err = -EUNIONFS_NO_WHITEOUT;

    print_entry_location();

    PASSERT(parent_dentry);
    old_bstart = dbstart(parent_dentry);
    old_bend = dbend(parent_dentry);

    fist_print_dentry("IN create_whiteout_parent", parent_dentry);

    /* create dentry's whiteout equivalent */
    name = KMALLOC(sizeof(char) * (strlen(filename) + 5), GFP_NOFS);
    if (!name) {
        err = -ENOMEM;
        goto out;
    }
    strcpy(name, ".wh.");
    strncat(name, filename, strlen(filename));
    name[strlen(filename) + 4] = '\0';

    for (bindex = start; bindex >= 0; bindex--) {

        hidden_parent_dentry = dtohd_index(parent_dentry, bindex);

	if (!hidden_parent_dentry) {
            /* create the recursive directory structure and return the
             * negative dentry for the parent where we want to create whiteout
             */
            ASSERT(parent_dentry->d_inode != NULL);
	    hidden_parent_dentry = unionfs_create_dirs(parent_dentry->d_inode, parent_dentry, bindex);
	    if (!hidden_parent_dentry || IS_ERR(hidden_parent_dentry)) {
		fist_dprint(8, "hidden dentry NULL for bindex = %d\n", bindex);
		continue;
	    } else {
	        /* create directory of the hidden parent, if it is negative.
                 * This is where whiteout is created
                 */
                hidden_grand_parent_dentry = lock_parent(hidden_parent_dentry);
                PASSERT(hidden_grand_parent_dentry->d_inode);

                /* We shouldn't create things in a read-only branch. */
                if (!(err = is_robranch_super(parent_dentry->d_sb, bindex))) {
                    err = vfs_mkdir(hidden_grand_parent_dentry->d_inode, hidden_parent_dentry, S_IRWXU);
                }

                unlock_dir(hidden_grand_parent_dentry);
                if (err || !hidden_parent_dentry->d_inode) {
		    dput(hidden_parent_dentry);
                    /* if error, we need to revert back the old bstart and bend for this
                     * parent and return back
                     */
                    set_dbstart(parent_dentry, old_bstart);
                    set_dbend(parent_dentry, old_bend);
                    set_dtohd_index(parent_dentry, bindex, NULL);

                    if (IS_SET(parent_dentry->d_sb, GLOBAL_ERR_PASSUP)) {
                        goto out;
                    }
                    /* break out of for loop if error returned was NOT -EROFS */
                    if (!IS_COPYUP_ERR(err)) {
                        break;
                    }

                    continue;
                }
                itohi_index(parent_dentry->d_inode, bindex) = igrab(hidden_parent_dentry->d_inode);
            }
        }

        /* lookup for the whiteout dentry that we want to create */
	hidden_wh_dentry = lookup_one_len(name, hidden_parent_dentry, strlen(filename) + 4);
	if (!hidden_wh_dentry || IS_ERR(hidden_wh_dentry)) {
	    continue;
	}
	ASSERT(!hidden_wh_dentry->d_inode);

        /* hidden_dir_dentry and hidden_parent_dentry are going to be same only */
	hidden_dir_dentry = lock_parent(hidden_wh_dentry);

        /* We shouldn't create things in a read-only branch. */
        if (!(err = is_robranch_super(parent_dentry->d_sb, bindex))) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
            err = vfs_create(hidden_dir_dentry->d_inode, hidden_wh_dentry, ~current->fs->umask & S_IRWXUGO);
#else
			err = vfs_create(hidden_dir_dentry->d_inode, hidden_wh_dentry, ~current->fs->umask & S_IRWXUGO,NULL);
#endif
	}

        unlock_dir(hidden_dir_dentry);
	dput(hidden_wh_dentry);

	if (!err) {
	    break;
        } else {
            if (IS_SET(parent_dentry->d_sb, GLOBAL_ERR_PASSUP)) {
                goto out;
            }
            /* break out of for loop if error returned was NOT -EROFS */
            if (!IS_COPYUP_ERR(err)) {
                break;
            }
        }
    } // end for

out:
    fist_print_dentry("OUT create_whiteout", parent_dentry);
    if (name) {
       KFREE(name);
    }
    print_exit_status(err);
    return err;
}

/* The callback structure for our whiteouts. */
struct unionfs_check_whiteout_callback {
    int err;
    int filldir_called;
    struct unionfs_dir_state *rdstate;
};

/* This is designed to be called just like a filldir function. */
static int check_whiteout(void *dirent, const char *name, int namelen, loff_t offset, ino_t ino, unsigned int d_type)
{
    int err = 0;
    struct unionfs_check_whiteout_callback *buf = (struct unionfs_check_whiteout_callback *) dirent;
    int whiteout = 0;
    struct filldir_node *found;

    print_entry_location();

    buf->filldir_called = 1;

    if (name[0] == '.' && (namelen == 1 || (name[1] == '.' && namelen == 2))) {
	goto out;
    }

    if (!strncmp(name, ".wh.", 4)) {
	namelen -= 4;
	name += 4;
	whiteout = 1;
    }

    found = find_filldir_node(buf->rdstate, name, namelen);
    /* If it was found in the table there was a previous whiteout. */
    if (found) {
	goto out;
    }

    /* If it wasn't found, and it isn't a whiteout the directory is not actually empty. */
    if (!whiteout) {
	err = -ENOTEMPTY;
	goto out;
    }

    err = add_filldir_node(buf->rdstate, name, namelen, buf->rdstate->uds_bindex, whiteout);

 out:
    buf->err = err;
    print_exit_status(err);
    return err;
}

// used by rmdir to check if a directory is indeed empty.
int check_empty(dentry_t *dentry, struct unionfs_dir_state **namelist)
{
    int err = 0;
    dentry_t *hidden_dentry = NULL;
    super_block_t *sb;
    file_t *hidden_file;
    struct unionfs_check_whiteout_callback *buf = NULL;
    int bindex, bstart, bend;

    print_entry_location();

    sb = dentry->d_sb;

    ASSERT(S_ISDIR(dentry->d_inode->i_mode));

    /* First we need to make sure all of the hidden_dentries are filled in. */
    err = unionfs_partial_lookup(dentry);
    if (err) {
	goto out;
    }

    bstart = dbstart(dentry);
    bend = dbend(dentry);

    buf = KMALLOC(sizeof(struct unionfs_check_whiteout_callback), GFP_NOFS);
    if (!buf) {
	err = -ENOMEM;
	goto out;
    }
    buf->err = 0;
    buf->rdstate = alloc_rdstate(dentry->d_inode, bstart);
    if (!buf->rdstate) {
	err = -ENOMEM;
	goto out;
    }

    /* Open all of the hidden_dentries and use the check_whiteout function as a filldir. */
    for (bindex = bstart; bindex <= bend; bindex++) {
	/* Find out hidden parent dentry */
	hidden_dentry = dtohd_index(dentry, bindex);

	/* continue if the hidden_dentry doesn't exist */
	if (!hidden_dentry) {
	    continue;
	}

	/* If it is a positive dentry. */
	if (hidden_dentry->d_inode && S_ISDIR(hidden_dentry->d_inode->i_mode)) {
	    dget(hidden_dentry);
	    mntget(stohiddenmnt_index(sb, bindex));
	    branchget(sb, bindex);
	    hidden_file = dentry_open(hidden_dentry, stohiddenmnt_index(sb, bindex), O_RDONLY);
	    if (IS_ERR(hidden_file)) {
		err = PTR_ERR(hidden_file);
		dput(hidden_dentry);
                branchput(sb, bindex);
		goto out;
	    }

	    do {
	        buf->filldir_called = 0;
		buf->rdstate->uds_bindex = bindex;
		err = vfs_readdir(hidden_file, check_whiteout, buf);
	        if (buf->err) {
		    err = buf->err;
	        }
	    } while (!err && buf->filldir_called);

	    /* fput calls dput for hidden_dentry */
	    fput(hidden_file);
            branchput(sb, bindex);

	    if (err < 0)
		goto out;
	}
    }

 out:
    if (buf) {
	if (namelist && !err) {
		*namelist = buf->rdstate;
	} else if (buf->rdstate) {
		free_rdstate(buf->rdstate);
	}
	KFREE(buf);
    }

    print_exit_status(err);
    return err;
}

// used by rmdir to nuke whiteouts in a directory
int delete_whiteouts(struct dentry *dentry, int bindex, struct unionfs_dir_state *namelist)
{
    int err = 0;
    struct dentry *hidden_dir_dentry = NULL;
    struct dentry *hidden_dentry;
    super_block_t *sb;
    char *name = NULL;

    int i;
    struct list_head *pos;
    struct filldir_node *cursor;

    print_entry_location();

    sb = dentry->d_sb;

    ASSERT(S_ISDIR(dentry->d_inode->i_mode));

    /* First we need to make sure all of the hidden_dentries are filled in. */
    err = unionfs_partial_lookup(dentry);
    if (err) {
	goto out;
    }

    ASSERT(bindex >= dbstart(dentry));
    ASSERT(bindex <= dbend(dentry));

    /* Find out hidden parent dentry */
    hidden_dir_dentry = dtohd_index(dentry, bindex);
    PASSERT(hidden_dir_dentry);
    PASSERT(hidden_dir_dentry->d_inode);
    ASSERT(S_ISDIR(hidden_dir_dentry->d_inode->i_mode));

    name = (char *)__get_free_page(GFP_NOFS);
    if (!name) {
	err = -ENOMEM;
	goto out;
    }

    for (i = 0; i < namelist->uds_size; i++) {
	list_for_each(pos, &namelist->uds_list[i]) {
	    cursor = list_entry(pos, struct filldir_node, file_list);
	    /* Only operate on whiteouts in this branch. */
	    if (cursor->bindex != bindex) continue;
	    if (!cursor->whiteout) continue;

	    strcpy(name, ".wh.");
	    strncpy(name + 4, cursor->name, PAGE_SIZE - 4);

	    hidden_dentry = lookup_one_len(name, hidden_dir_dentry, cursor->namelen + 4);
	    if (IS_ERR(hidden_dentry)) {
		err = PTR_ERR(hidden_dentry);
		goto out;
	    }
	    if (!hidden_dentry->d_inode) {
		dput(hidden_dentry);
		continue;
	    }

	    down(&hidden_dir_dentry->d_inode->i_sem);
	    err = vfs_unlink(hidden_dir_dentry->d_inode, hidden_dentry);
	    up(&hidden_dir_dentry->d_inode->i_sem);
	    dput(hidden_dentry);

	    if (!err) {
		d_delete(hidden_dentry);
	    } else if (!IS_COPYUP_ERR(err)) {
		goto out;
	    }
	}
    }

 out:
    /* After all of the removals, we should copy the attributes once. */
    fist_copy_attr_times(dentry->d_inode, hidden_dir_dentry->d_inode);
    dentry->d_inode->i_nlink = get_nlinks(dentry->d_inode);

    if (name) {
	free_page((unsigned long) name);
    }
    print_exit_status(err);
    return err;
}

/* This is a helper function for rename, which ends up with hosed over dentries
 * when it needs to revert. */
int unionfs_refresh_hidden_dentry(struct dentry *dentry, int bindex) {
    struct dentry *hidden_dentry;
    struct dentry *hidden_parent;
    int err = 0;

    print_entry(" bindex = %d", bindex);

    hidden_parent = dtohd_index(dentry->d_parent, bindex);

    PASSERT(hidden_parent);
    PASSERT(hidden_parent->d_inode);
    ASSERT(S_ISDIR(hidden_parent->d_inode->i_mode));

    hidden_dentry = lookup_one_len(dentry->d_name.name, hidden_parent, dentry->d_name.len);
    if (IS_ERR(hidden_dentry)) {
	err = PTR_ERR(hidden_dentry);
	goto out;
    }

    if (dtohd_index(dentry, bindex)) {
    	dput(dtohd_index(dentry, bindex));
    }
    if(itohi_index(dentry->d_inode, bindex)) {
    	iput(itohi_index(dentry->d_inode, bindex));
	itohi_index(dentry->d_inode, bindex) =  NULL;
    }
    if (!hidden_dentry->d_inode) {
	dput(hidden_dentry);
    	set_dtohd_index(dentry, bindex, NULL);
    } else {
    	set_dtohd_index(dentry, bindex, hidden_dentry);
    	itohi_index(dentry->d_inode, bindex) = igrab(hidden_dentry->d_inode);
    }

out:
    print_exit_status(err);
    return err;
}

/*
 * vim:shiftwidth=4
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
